"""Utilities"""
import string
import binascii


def escape_nonprintables(input_string):
    """Replace non-printable characters and line breaks in a string
       with escaped hex codes."""

    result = ''
    printable = string.ascii_letters + string.digits + string.punctuation + ' '
    for char in input_string:
        if char not in printable:
            result += '\\x' + binascii.hexlify(char)
        else:
            result += char
    return result


def contains_nonprintables(input_string):
    """Check if non-printable characters are in the string."""

    for char in input_string:
        if char not in string.printable:
            return True
    return False



"""Classical cryptography"""
import string


def letter_to_num(letter):
    """Return the position of letter in the English alphabet."""

    if not isinstance(letter, str):
        raise TypeError("Expected a string")

    if letter not in string.ascii_letters:
        raise ValueError("Expected a single ASCII letter")

    return ord(letter.upper()) - ord('A') + 1


def num_to_letter(num):
    """Return the letter of the English alphabet at the given position."""

    if not isinstance(num, int):
        raise TypeError("Expected an integer")

    if num < 1 or num > 26:
        raise ValueError("Expected an integer between 1 and 26 (inclusive)")

    return chr(num-1 + ord('A'))


def caesar_shift_letter(letter, shift):
    """Return the Caesar shifted letter."""

    if shift < 0:
        raise ValueError("Expected a positive shift value")
    shift %= 26

    num = letter_to_num(letter) - 1
    num = ((num + shift) % 26) + 1

    return num_to_letter(num)


def caesar_shift(text, shift):
    """Return the Caesar shifted text."""

    result = ""
    for i in range(len(text)):
        if text[i] in string.ascii_letters:
            result += caesar_shift_letter(text[i], shift)
        else:
            result += text[i]
    return result



"""Frequency analysis, etc."""
import math
from .frequencies import ENGLISH_FREQ

# Non-alphabetic ASCII characters
NON_ALPHAS = ''.join(c for c in map(chr, range(256)) if not c.isalpha())


def char_frequency(text, only_alpha=False):
    """Return a table of character frequencies in the text.
       Case insensitive; includes non-alphabetic characters
       by default."""

    if only_alpha:
        text = text.translate(None, NON_ALPHAS)

    text_count = {}
    for char in text:
        if char.lower() in text_count:
            text_count[char.lower()] += 1
        else:
            text_count[char.lower()] = 1

    # Get frequency for each character counted
    text_freq = {x[0]: float(x[1])/len(text) for x in text_count.items()}

    return text_freq


def cosine_similarity(left_vector, right_vector):
    """Compute cosine similarity of two vectors.

    (v1 dot v1)/{||v1||*||v2||)
    http://stackoverflow.com/questions/18424228"""

    sumxx, sumxy, sumyy = 0, 0, 0
    for i in range(len(left_vector)):
        x = left_vector[i]
        y = right_vector[i]
        sumxx += x*x
        sumyy += y*y
        sumxy += x*y

    if sumxx == 0 or sumyy == 0:
        return float(0)

    return float(sumxy)/math.sqrt(sumxx*sumyy)


def score_text(text, freq=ENGLISH_FREQ):
    """Score how well a text matches a given frequency table.
    Case insensitive. Uses English letter frequency by default.
    Only characters with an entry in the frequency table are considered.
    The result is the cosine similarity of the frequency tables of the text
    and the language, where a value of 1 indicates most similar and
    a value of -1 indicates least similar."""

    if len(text) == 0:
        return 0
    if len(freq) == 0:
        raise Exception("Empty frequency table")

    for k in freq.keys():
        if k != k.lower():
            raise ValueError("Only use lower-case keys for frequency table")

    text_freq = char_frequency(text)

    # Get vector for each frequency table
    text_vector = []
    freq_vector = []
    for k in freq.keys():
        freq_vector.append(freq.get(k))
        text_vector.append(text_freq.get(k) or 0.0)

    return cosine_similarity(text_vector, freq_vector)


def hamming_distance(left_bytes, right_bytes):
    """Get the binary Hamming distance between two sequences of bytes."""

    if len(left_bytes) != len(right_bytes):
        raise Exception("Length of both sequences must match")

    result = 0
    for i in range(len(left_bytes)):
        # Each 1 bit in the XOR represents a difference
        diff = ord(left_bytes[i]) ^ ord(right_bytes[i])
        for j in range(8):
            result += (diff >> j) & 1
    return result



"""Bits & bytes"""
import base64
import binascii


def hex_to_bytes(hex_string):
    """Convert a string of hexadecimal values into bytes."""

    return binascii.unhexlify(hex_string)


def bytes_to_hex(byte_string):
    """Convert a string of bytes into hexadecimal string representation."""

    return binascii.hexlify(byte_string)


def bytes_to_b64(input_bytes):
    """Return base64 encoding of bytes."""

    return base64.b64encode(input_bytes)


def hex_to_b64(hex_string):
    """Return base64 encoding of bytes given in hexadecimal form."""

    return bytes_to_b64(hex_to_bytes(hex_string))


def xor_bytes(input_bytes, key_bytes):
    """XOR input bytes, repeating the XOR key as necessary."""

    result = ''
    for i in range(len(input_bytes)):
        result += chr(ord(input_bytes[i]) ^ ord(key_bytes[i % len(key_bytes)]))
    return result



import os
from setuptools import setup

long_description = open(
    os.path.join(os.path.dirname(__file__), 'README.rst')).read()

setup(
    name='etao',
    version='0.2.1',

    description='Simple cryptanalysis library',
    long_description=long_description,

    url='https://github.com/jamchamb/etao',

    author='James Chambers',
    author_email='jameschambers2@gmail.com',

    license='GPLv3',

    keywords=[
        'cryptography',
        'cryptanalysis',
        'ciphers'
    ],

    test_suite='nose2.collector.collector',

    packages=['etao'],
    install_requires=[
        'nose2>=0.6.2',
        'six>=1.10.0'
    ],
    zip_safe=False
)



"""General cryptography/cryptanalysis functions."""

# Disable "unused imports" nags
# flake8: noqa

from .pen import (letter_to_num, num_to_letter,
                  caesar_shift_letter, caesar_shift)
from .analysis import char_frequency, score_text, hamming_distance
from .bitwise import (bytes_to_hex, hex_to_bytes, bytes_to_b64,
                      hex_to_b64, xor_bytes)
from .util import escape_nonprintables, contains_nonprintables
